; devramx.asm ( ASSEMBLER SOURCE CODE MODULE )
; faked device file for use with truss.asm
;
; Hooks some of the interrupt 21H services - reading/writing and other
; handling of open files - in order to treat a portion of ram as a file
; The ram in question is mapped to the label ram: at the end of this
; module. If XMS is available then extended memory is used instead.
;
; Access to the ramfile must be triggered by an actual file-system reference
; in order to obtain a valid handle.  
;
; This file should be assembled and linked entirely by itself and then
; converted to a memory image with exe2bin. If the result is called 
; devramx.bin, then truss devramx.bin [args ...] will install a tsr with
; device.bin attached to it, and then run the installation code of this
; module before exiting. 

.MODEL tiny

_TEXT SEGMENT PUBLIC 'CODE'
ASSUME cs:_TEXT,ds:_TEXT
PUBLIC 
EXTRN
HOOK equ 21H      ; interrupt to listen for
FRAME_BYTES equ 16384 ; size of ramfile that is in conventional memory - as
		      ; opposed to XMS. It is the default size unless more is
		      ; from main. This should match DEFAULT_PARS * 16
		      ; This is 2^14
LO_FRAME_MASK equ 3FFFH ; This is 2^14 - 1
HI_FRAME_MASK equ 0C000H ; This is 2^16 - 2^14 
 

org 00h         ; this code called after having been copied to the end
                ; of the resident truss code at offset 400H = 40H par 
		; Entry point is at offset 0, unlike .coms, which start
		; at 100h

jmp os_cionn    ; These two need to stay unchanged. In part,
org 03h         ; they constitute a 3 byte "magic number"
jmp _unhook     ; that is checked by the loader in main.c  
os_cionn:
jmp initialize_device
;;;;;;;;;;;;;;;
; device data ;
;;;;;;;;;;;;;;;
device_name db 'C:\0',0,0,0,0,0,0,0,0,0,0,0,0  ; 16 bits for C:\filename.ext
device_name_length dw (0005H)       ; must include null character at end
file_name_seg dw ?
file_name_off dw ?
old_handler_seg dw ?      ; address of old 21H handler
old_handler_off dw ?
tsr_psp_seg dw ?
caller_psp_seg dw ?
little_buf db 'xxxxxxxx.xxx'
quiet_flag db 0
name_flag db (0)         ; flag to skip read-only test
opened db (0)             ; flag for whether we are opened or not
written db (0)            ; flag for whether file has been written to
; the C library wants the file ptr to be an unsigned long
; but we don't currently use more than the low word
file_ptr dd (00000000H)       ; simulated file pointer 
max_file_ptr dd ?
end_of_file dd (00000000H)
dos_handle dw (00FFH)     ; msdos handle for the device place-holder
file_access db (02H)      ; file access, rw is assumed unless changed by open
XMSSVR_OFF dw ?
XMSSVR_SEG dw ?
_XMSversion dw ?
_HIMEMversion dw ?
_xms_avail dw ?
_tot_xms  dw ?
_xms_handle dw ?
_xms_mem_req dw 400H      ; 1024k = 1 meg 
using_xms db 00H          ; flag
my_greeting db 'ramfile (1.0) Chicken Haven Software',0DH,0AH,'$'
exit_msg db 'ramfile: unhooking interrupt vector',0DH,0AH,'$'
error_msg db 'There is something wrong',0DH,0AH,'$'
NO_XMSSERVER_MSG db 'No XMS server detected',0DH,0AH,'$'
NO_XMS_HANDLE_MSG db 'Unable to obtain XMS handle ',0DH,0AH,'$'

_unhook:                  ; restore original handler for HOOK
			  ; Called from uninst resident truss code
	push ax
	push ds
	push dx
	mov ax,cs         ; fake out ds to reference seg address of start
	add ax,40H	  ; 
	mov ds,ax

	; release xms memory if it is being used
	mov al,using_xms
	cmp al,0
	je no_xms_to_free
	mov ah,0AH       ; free extended memory block
	mov dx,_xms_handle
	call DWORD PTR XMSSVR_OFF 

no_xms_to_free:
;	mov dx, offset exit_msg
;	mov ah, 09H          ; Display greeting
;	int 21H

	mov ax, old_handler_off
	mov dx,ax
	mov ax, old_handler_seg
	mov ds,ax
	mov al, HOOK
	mov ah, 25H        ; set interrupt vector
	int 21H	
	pop dx
	pop ds
	pop ax
	ret

error_handler:
	mov dx, offset error_msg
	mov ah, 09H          ; Display greeting
	int 21H
	jmp go_home

	
; far *initialize_device(unsigned char, unsigned char, unsigned) 
initialize_device:            ; Program entry point label 
; Initialization code: Hook interrupt 21H services to create ramfile
	
	push bp               ; set up stack frame
	mov bp,sp

			  ; preserve registers changed in this call
	                  ; using stack of main when we are called
	push bx
	push cx
	push es
	push dx
	push ds

	mov ax,cs	  ; adjust ds to point to start 
	add ax,40H        ; of this module, which is at offset 40H in tsr
	mov ds,ax

	mov ax,cs         ; seg address of tsr psp
	mov word ptr tsr_psp_seg,ax

	; Danger, Will Robinson. Accessing arguments passed by C compiler
        ; is tricky, because it is impossible to predict with certainty
        ; how the compiler will do it. Borland C promotes chars to ints
	; when they are passed.
	mov ax,[bp + 10]       ; access unsigned arg, 
	mov word ptr _xms_mem_req,ax  ; in k bytes  
	mov al,[bp + 8]         ; access 2nd char arg 
	mov byte ptr quiet_flag,al ;
	mov al,[bp + 6]         ; access unsigned char arg (passed as int)
	mov byte ptr name_flag,al;
	mov word ptr end_of_file, 0000H
	mov word ptr end_of_file[2], 0000H
	mov word ptr file_ptr, 0000H
	mov word ptr file_ptr[2], 0000H

; get and store address of current handler

	mov al, HOOK
	mov ah, 35H       ; get interrupt vector
	int 21H
	mov ax,es
	mov old_handler_seg,ax
	mov old_handler_off,bx

;	mov dx, offset my_greeting
;	mov ah, 09H          ; Display greeting
;	int 21H

; check to see if an XMS server V2+ installed

	mov ax,4300H    ; Int 2F service XMS Installation Check
	int 2FH
	cmp al,80H      ; Succeeds if service available
	je  xms_ok      
	mov dx, FRAME_BYTES
	mov WORD PTR max_file_ptr, dx 
	xor dx,dx
	mov WORD PTR max_file_ptr[2],dx
	mov dx,offset NO_XMSSERVER_MSG
	mov ah,09H
	int 21H
	jmp go_home

MEM_AVAIL_MSG db 'Available XMS memory (KB) = ','$'
MEM_TOTAL_MSG db ' Total = ','$'
NEW_LINE db 0AH,0DH,'$'

xms_ok:
	; Record that we are using xms so server address is known to be valid
	mov BYTE PTR using_xms, 01H
	; Get the server address and store in XMSSVR_SEG and XMSSVR_OFF
	mov ax, 4310H
	int 2FH        ; ES:BX now has server address
	mov ax,ES
	mov XMSSVR_SEG, ax
	mov XMSSVR_OFF,bx

	; The following gives the paradigm for calling on the server

	; Get available memory
	mov ah, 08H     ; Query free extended memory
	mov bl, 00h     ; XMS function returns error code in bl
	call DWORD PTR XMSSVR_OFF 

	mov _xms_avail,ax
	mov _tot_xms,dx

	; Get the version numbers
	xor ah,ah
	call DWORD PTR XMSSVR_OFF
	mov _XMSversion,ax
	mov _HIMEMversion,bx

	cmp quiet_flag,0H 
	jne skip_report
	; report version and other numbers
	call print_version
	mov dx, offset MEM_AVAIL_MSG
	mov ah,09H
	int 21H
	mov ax,_xms_avail
	call print_num
	mov dx,offset MEM_TOTAL_MSG
	mov ah,09H
	int 21H
	mov ax,_tot_xms
	call print_num
	mov dx,offset NEW_LINE
	mov ah,09H
	int 21H
skip_report:

; Get an XMS handle
	mov ah,09H        ; Allocate Extended Memory Block
	mov dx,_xms_mem_req
	cmp dx,0
	je got_no_xms_handle 
	call DWORD PTR XMSSVR_OFF
	cmp ax,00H        ; did it succeed?
	je  got_no_xms_handle 
	mov _xms_handle,dx
	jmp go_home

got_no_xms_handle:
	mov dx, FRAME_BYTES
	mov WORD PTR max_file_ptr, dx 
	xor dx,dx
	mov WORD PTR max_file_ptr[2],dx
	mov dx, offset NO_XMS_HANDLE_MSG ; Bug: we should also report
	mov ah,09H                       ; error code in BL
	int 21H
	mov BYTE PTR using_xms, 00H
	
go_home:                   

; point vector to us

	mov ax, offset new_handler
	mov dx,ax
	mov al, HOOK
	mov ah, 25H        ; set interrupt vector
	int 21H	

; return to startup truss code leaving this stuff
	pop ds             ; resident
	pop dx
	pop es
	pop cx
	pop bx
	pop bp 

	retf               ; far return to code that started us up	

new_handler:               ; Get here on int21H
			   ; Unlike initialization code, when we come here
	                   ; cs points to the beginning of this module
			   ; and ds points who knows where?
			   ; switch stacks

	pushf              ; save flag values on entry since cmp's change them

	cmp byte ptr cs:opened,00H
	je case_closed     ; what to do when the ramfile is closed ...

	; if the ramfile is open, we only take action on services
	; that supply our own stored handle in bx

	cmp cs:dos_handle,bx  ; all these calls pass handle in bx
	jne orig_vector   ; quick and dirty exit so as not to bog down 21H

;call debug

	; switch to service selected by ah - ones we handle, anyway

	cmp ah,3EH
 	jne eee
	call case_3EH
	jmp common_iret
eee:
	cmp ah,3FH
	jne bbb
	call case_3FH
	jmp common_iret
bbb:
	cmp ah,40H
	jne ccc
	call case_40H
	jmp common_iret
ccc:
	cmp ah,42H
	jne ddd 
	call case_42H
	jmp common_iret

ddd: 
	cmp ax,4400H
	jne fff
	call case_4400H
	jmp common_iret

fff:  
	cmp ax,440AH
	jne ggg
	call case_4400H  ; for us, this is the same as 4400H
	jmp common_iret

ggg:
	cmp ax,4406H
	jne iii
	call case_4406H
	jmp common_iret

iii:
	cmp ah,3DH
	je case_closed    ; not really, but we handle two opens there
	jmp orig_vector   ; not one of the services we care about  

common_iret:
	jnc normal_iret
	; carry indicates error condition. 
	; ERROR_INVALID_HANDLE can never occur for us
	popf
	stc
	retf 2
	
normal_iret:
	popf  ; restore flags pushed at start  
	clc
	retf 2

; Go to original vector and don't return here
orig_vector:
	popf                 ; restore flags on entry
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; File system services    ;;
;; Called when ramfile is  ;;
;; closed, or when file    ;;
;; name rather than handle ;;
;; is passed               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_closed: 

; if the ramfile is closed we only take action on open file 

	cmp ah,3DH
	jne orig_vector
	
; save stuff 
	mov byte ptr cs:file_access,al  ; save for later use
	mov word ptr cs:file_name_off,dx 
	mov ax,ds
	mov word ptr cs:file_name_seg,ax

	push ds
	push es
	push dx
	push bx
	push cx

; see if we are going to go by our stored mount point name (-m option)
; or to allow any read-only zero size file as mount point

	mov al,byte ptr cs:name_flag
	cmp al,00H
	je read_only_zero_size

	mov ax,ds
	mov es,ax    ; es:dx points at open file name	
	mov ax,cs    ; load ds 
	mov ds,ax
	mov bx,dx    ; es:bx points at open file name
	mov dx, offset device_name  ; ds:dx points at our name
	mov cx, word ptr device_name_length
	call strncmp ; compare open file name with our name
	cmp ax,0000H
	je ok_to_open2 ; same, so open and get on with life
	jmp open_with_prefs ; not the same, so open with original access req

read_only_zero_size:

; Now we call the original int 21H handler, but return here
; This same code is used 4 different places, but in-lined for speed

	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again
	push ax
	mov al,00H ; open read only for testing
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again:
	jc open_with_prefs  ; try to open with original access request 

	mov bx,ax    ; handle returned from open in bx

	; See if it is a device rather than a file
	; If it were, e.g., CON, the read test for zero length below
        ; would hang waiting for input, so we don't want to do that
	; 

	mov ax,4400H ; Get device data
	int 21H	
	cmp dx,128
	jb its_a_file    ; it's a device if bit 7 is set (= 128)
	mov ah,3EH       ; close device
	int 21H
	jmp open_with_prefs

ok_to_open2:
jmp ok_to_open           ; dratted 128 byte limit!

its_a_file:

	; else, it's a file
	; try to read one byte from the open file. If it succeeds,
	; iret to whomever wanted the handle -
        ; not us, since we only deal with place-holder files of zero length

	mov cx,0001H ; ask to read one byte
	mov ax,cs    ; load ds for this segment
	mov ds,ax
	mov dx, offset little_buf   ; ds:dx = our little buffer
	mov ah,3FH     ; read file or device
	int 21H
	cmp ax,0000H   ; see how many bytes were read
	je zero_size_file  ; if none, then we are good to go so far

	mov ah,3EH ; close file
	int 21H
	jmp open_with_prefs
	
zero_size_file:

	; close the file and try to open for write. If this fails
	; then we are definitely good to go - file is read-only
	mov ah,3EH   ; close file
	int 21H

	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again2
	push ax
	mov al,02H ; rw for testing
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again2:
	jc ok_to_open

	mov ah,3EH   ; close file
	int 21H

open_with_prefs:

	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again4
	push ax
	mov al,byte ptr cs:file_access ; stored original parameters
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again4:
	jc error_return
	jmp good_return 
	
ok_to_open:
	; open place-holder file for good
	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again3
	push ax
	mov al,00H ; open place-holder file read-only
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again3:
	mov word ptr cs:dos_handle,ax      ; store handle
	mov byte ptr cs:opened,01H          ; mark the ramfile as open
	mov word ptr cs:file_ptr[2], 0000H  ; zero-out file pointer
	mov word ptr cs:file_ptr, 0000H ; 
	mov byte ptr cs:written, 00H        ; zero-out dirty file flag

good_return:

	pop cx
	pop bx
	pop dx
	pop es
	pop ds
	;  handle is returned in ax 
	popf ; restore flags 
	retf 2   ; hard lesson: don't iret here because it will
                 ; popf and obliterate any error return

error_return:

	pop cx
	pop bx
	pop dx
	pop es
	pop ds
	; error is returned in ax 
	popf ; restore flags 
	stc  ; but set carry to indicate error
	retf 2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Handle Services - called ;;
;; When ramfile is open     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_42H proc
   ; move file pointer
	; cx-dx has hi-lo words of displacement (signed 32 bit)
	; al has method 00 = from start, 01 = from current, 02 = from end
	cmp al,00H
	je from_start
	cmp al,01H
	je from_current
	cmp al,02H
	je from_end
	mov ax, 0001H ; error ERROR_INVALID_FUNCTION
	stc
	ret
from_start:
	mov word ptr cs:file_ptr[2],cx
	mov word ptr cs:file_ptr,dx
	jmp done42
from_current:
	add word ptr cs:file_ptr,dx    ; N.B. move past eof is possible
	adc word ptr cs:file_ptr[2],cx       
	jmp done42
from_end: 
	add dx,word ptr cs:end_of_file
	adc cx,word ptr cs:end_of_file[2]
	mov word ptr cs:file_ptr,dx
	mov word ptr cs:file_ptr[2],cx
done42:
	mov ax, word ptr cs:file_ptr
	mov dx, word ptr cs:file_ptr[2]
	ret
endp

case_3EH  proc 

; close file with handle 
	push ax
	push ds
	push cx
	push dx

	mov ax,cs
	mov ds,ax
	mov byte ptr opened,00H       ; mark ram-file as closed
				      ; doing here makes sure int 21H
                                      ; calls below do not recurse
	mov word ptr dos_handle,00FFH ; restore dos_handle to invalid value

	cmp byte ptr written,00H   ; has the file been modified?
	je ok_to_close             ; no, so just go ahead and close

	; Need to time stamp place-holder file. Get current date
	mov ah,2AH
	int 21H
		     ; this puts day of month in dl, month number 1,...12
	             ; in dh, and year number in cx (from 1980)
		     ; need to repackage this into one register	
	push bx      ; store because we need bx
	sub cx, 1980
 	xor bx,bx
	mov bl,dh 
	sal bl,1
	sal bl,1
	sal bl,1
	sal bl,1
	sal bl,1
	add bl,dl
	adc bh,cl    ; bx should now have date that 5701H wants in dx
	 
	mov ah,2CH
	int 21H      ; Get Time
		     ; military hour in ch, minutes in cl, seconds in dh
	mov dl,ch
	xor ch,ch
	sal cx,1
	sal cx,1
	sal cx,1
	sal cx,1
	sal cx,1
	shr dh,1
	add cl,dh
	xor dh,dh
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	add cx,dx    ; cx now has time format 5701H wants
	mov dx,bx    ; now dx had date 
	pop bx       ; handle back in bx
	mov ax,5701H
	int 21H      ; set file date and time

ok_to_close:
	mov ah,3EH   ; Actually ask Dos to close handle of placeholder file
	int 21H      ; Will not recurse because ramfile now closed

	pop dx
	pop cx
	pop ds
	pop ax
	ret          ; return possibly with carry flag set on error
endp

case_4400H proc
	; Get device data
	; or 440A, is file or device remote?  (no, actually)
	cmp byte ptr cs:written,00H
	je not_written
	mov dx,0002H     ; written, and disk = 2 (C:)
	ret
not_written:
	mov dx,0042H     ; not written, and disk = 2
	ret
endp

case_4406H proc
	; Is device ready for input ?
	; or 4407H, for output ?
	mov al,0FFH ; yes, it is
	ret

endp

; Structure used to move blocks to and from extended memory
; Set Handle to 0 when referencing conventional memory
; in which case offset is segment:offset pair, but bass-ackwards: segment
; is low byte and offset is high byte - go figure!

ExtMemMoveStruct struc
 Lngth          dd ? ; 32-bit number of bytes to transfer
 SourceHandle    dw ? ; Handle of source block
 SourceOffset    dd ? ; 32-bit offset into source
 DestHandle      dw  ? ; Handle of destination block
 DestOffset      dd  ? ; 32-bit offset into destination
ExtMemMoveStruct ends

myxms ExtMemMoveStruct <0,,,,,> ; dos system calls are limited to moving 64k
                                ; bytes at a time, so we make sure high byte
                                ; is initialized to 0

; Copy frame buffer to extended memory. Destination frame is the one
; containing current value of file ptr  
; Does no bounds checking or reset of file ptr

write_frame proc

	push ax
	push bx
	push cx
	push dx
	push ds                     ; save registers
	push si
	mov ax,cs
	mov ds,ax                   ; load ds for our data segment
	xor ax,ax
	mov myxms.SourceHandle,ax   ; mark as source in conventional mem 
	mov Word Ptr myxms.SourceOffset[0],offset ram ; here we give
	mov Word PTR myxms.SourceOffset[2],ds         ; seg:offset pair
;	mov Word Ptr myxms.SourceOffset[2],offset ram ; here we give
;	mov Word PTR myxms.SourceOffset[0],ds         ; seg:offset pair
	mov ax,_xms_handle   
	mov myxms.DestHandle,ax   ; handle to our xms mem
	mov ax,word ptr file_ptr[0]
	and ax,HI_FRAME_MASK        ; round to (next below) nearest 
			            ; FRAME_BYTES bytes. Write behind
				    
	mov word ptr myxms.DestOffset[0],ax
	mov ax,word ptr file_ptr[2]
	mov word ptr myxms.DestOffset[2],ax
	mov WORD PTR myxms.Lngth[0],FRAME_BYTES 
	mov Word PTR myxms.Lngth[2],00H
	mov si,offset myxms            ; ds:si -> move struct
	mov ah, 0Bh                    ; Move extended mem block
	call DWORD PTR XMSSVR_OFF      ; via xms server call
	cmp ax,00H
	jnz no_xms_write_error         ; success when ax not zero
	mov byte ptr xms_write_message,bl
	mov dx, offset xms_write_message 
	mov ah, 09H          ; Display greeting
	int 21H
no_xms_write_error:
	pop si                         ; restore registers
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	ret
endp

xms_write_message db '  = error writing xms memory',0DH,0AH,'$'

; Bring in frame buffer full from extended memory
; to frame buffer, beginning at current value of file_ptr 
; (which should be congruent to zero modulo FRAME_BYTES)
; Does no bounds checking, no reset of file ptr

load_frame proc

	push ax
	push bx
	push cx
	push dx
	push ds                   ; save registers
	push si
	mov ax,cs                 ; load ds with our data segment
	mov ds,ax
	xor ax,ax
	mov myxms.DestHandle,ax   ; mark destination as in conventional mem 
	mov Word Ptr myxms.DestOffset[0],offset ram  ; dest is our frame
	mov Word PTR myxms.DestOffset[2],ds          ; buffer
	;mov Word Ptr myxms.DestOffset[2],offset ram  ; dest is our frame
	;mov Word PTR myxms.DestOffset[0],ds          ; buffer
	mov ax,_xms_handle   
	mov myxms.SourceHandle,ax   ; handle to our xms mem
	mov ax,word ptr file_ptr[0] ; source is mblock beginning at file ptr
	mov word ptr myxms.SourceOffset[0],ax
	mov ax,word ptr file_ptr[2]
	mov word ptr myxms.SourceOffset[2],ax
	mov WORD PTR myxms.Lngth[0], FRAME_BYTES 
	mov WORD PTR myxms.Lngth[2], 00H
	mov si,offset myxms            ; ds:si -> move struct
	mov ah, 0Bh                    ; Move extended mem block
	call DWORD PTR XMSSVR_OFF      ; by calling xms server
	cmp ax,00H
	jnz no_xms_read_error
	mov byte ptr xms_read_message,bl
	mov dx, offset xms_read_message 
	mov ah, 09H          ; Display greeting
	int 21H
no_xms_read_error:
	pop si                         ; restore registers
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	ret
endp

xms_read_message db ' = xms read error',0DH,0AH,'$'

case_3FH proc
   ; read file or device
   ; on entry ds:dx points to caller buffer to read into
   ; and cx has number of bytes to read

	cmp byte ptr cs:file_access,01H
	jne ok_to_read
	mov ax,0005H ; ERROR_ACCESS_DENIED
	stc          ; flag error
	ret

ok_to_read:

	push ds         ; save registers 
	push si
	push es
	push di
	push dx
	push bx

	mov di,dx
	mov ax,ds
	mov es,ax       ; es:di now points to caller buffer to read into
	mov ax,cs
	mov ds,ax       ; load ds -- point to this data segment

	mov bx,00H      ; track actual copied bytes in bx for return value
			; make sure there is room before eof to satisfy 
                        ; request, otherwise shorten request
	mov dx, cx      ; save original bytes to copy in dx	
	mov ax, word ptr end_of_file[2];
	sub ax, word ptr file_ptr[2];
	cmp ax,0
	ja refill_loop  ; there is clearly enough space since dos read requests
                        ; can never exceed 64k
	mov ax, word ptr end_of_file[0]
	sub ax, word ptr file_ptr[0]
	cmp ax,cx
	ja refill_loop ; there is enough space
	mov dx,ax      ; adjust request if not. On return, this indicates eof	

refill_loop:
	; reduce file ptr modulo FRAME_BYTES = bytes in frame buffer
	mov ax,word ptr file_ptr[0] ; lo byte sufficient for div
	and ax,LO_FRAME_MASK        ; ax = file_ptr mod FRAME_BYTES
	cmp ax,0
	ja skip_load    ; no need to refill frame: file ptr is not at end 
	cmp using_xms,0
	je skip_load
	call load_frame ; refill frame buffer from xms
skip_load:
	mov cx, FRAME_BYTES
	sub cx, ax      ; cx has bytes in frame from pointer onwards
	cmp cx, dx      ; are there enough bytes remaining in frame for copy?
	jb no_trunc     ; yes, don't adjust number to copy
	mov cx,dx       ; else copy only up to request

no_trunc:
	mov si,offset ram                ; ready to read next frame
	add si,ax                        ; aim si
	mov ax,cx                        ; preserve copy number
	cld
	rep movsb                        ; copy to caller buffer
	add bx,ax                        ; update actual copied bytes
	add word ptr file_ptr[0],ax         ; advance 32 bit file ptr
	adc word ptr file_ptr[2],0H
	sub dx,ax                        ; decrement request
	cmp dx,0                         ; and quit loop if request filled
	ja refill_loop
	mov ax,bx       ; return in ax actual bytes copied
	clc             ; indicate no error
	pop bx          ; restore registers
	pop dx
	pop di
	pop es
	pop si
	pop ds
	ret
endp

case_40H proc

 ; write file or device

	     ; on entry ds:dx points to buffer containing data
	     ; cx has requested bytes to write (if possible)

	cmp byte ptr cs:file_access,00H ; read-only
	jne ok_to_write
	mov ax,0005H ; ERROR_ACCESS_DENIED
	stc
	ret

ok_to_write:

	push ds         ; preserve registers used
	push dx
	push si
	push es
	push di
	push bx

	mov si,dx       ; ds:si points to buffer to write from
	mov ax,cs
	mov es,ax

			; note that we are stuck with caller's ds,
			; so segment over-ride must be used on all
			; memory accesses to this (cs) segment 


	mov bx,00H      ; track actual copied bytes in bx for return value
			; make sure there is room to satisfy 
                        ; request, otherwise shorten request
	cmp cs:using_xms,00H
	je use_frame_eof
	mov ax,400H
	mul word ptr cs:_xms_mem_req ; convert ram file in xms size to bytes
	sub dx, word ptr cs:end_of_file[2];
	cmp dx,0
	ja enough_space  ; there is clearly enough space since dos write requests
                        ; can never exceed 64k
	sub ax, word ptr cs:end_of_file[0]
	cmp ax,cx
	ja enough_space ; there is enough space before eof
	mov cx,ax      ; adjust request if not. On return, this indicates eof	
	jmp enough_space

use_frame_eof:
	mov ax,word ptr cs:max_file_ptr[0]
	sub ax, word ptr cs:end_of_file[0]
	cmp ax,cx
	ja enough_space
	mov cx,ax

enough_space:
	mov dx,cx      ; save adjusted request size 

refill_loop2:

	; reduce file ptr modulo FRAME_BYTES = bytes in frame buffer
	mov ax,word ptr cs:file_ptr[0] ; lo byte sufficient for div
	and ax,LO_FRAME_MASK           ; ax = file_ptr mod FRAME_BYTES
	mov cx, FRAME_BYTES
	sub cx, ax      ; cx has bytes in frame from pointer onwards
	cmp cx, dx      ; are there enough bytes remaining in frame for copy?
	jb no_trunc2     ; yes, don't adjust number to copy
	mov cx,dx       ; else copy only up to request

no_trunc2:
	mov di,offset ram                ; ready to read next frame
	add di,ax                        ; aim di
	mov ax,cx                        ; preserve copy number
	cld
	rep movsb                        ; copy from caller buffer
   	mov byte ptr cs:written,01H
	add bx,ax                        ; update actual copied bytes
	sub dx,ax                        ; decrement request
	cmp cs:using_xms,0
	je skip_write
	call write_frame                 ; write-behind update ramfile in xms
skip_write:
	add word ptr cs:file_ptr[0],ax
	adc word ptr cs:file_ptr[2],0H
	mov ax, word ptr cs:file_ptr[0]
	mov word ptr cs:end_of_file[0], ax
	mov ax, word ptr cs:file_ptr[2]
	mov word ptr cs:end_of_file[2],ax
	cmp dx,0                         ; and quit loop if no request left
	ja refill_loop2

	mov ax,bx       ; return actual bytes copied in ax
	clc             ; signal that no error occurred
	pop bx          ; restore registers
	pop di
	pop es
	pop si
	pop dx
	pop ds
	ret
endp
debug_msg db '  (int 21H service) ',0DH,0AH,'$'
service_code db ?
sub_code db ?

debug proc

	push ds
	push bx
	push ax
	push dx
	mov cs:service_code,ah
	mov cs:sub_code,al
	add cs:sub_code,20H ; make readable by moving into printable range

	mov bx,0FFFH    ; prevent recurse
	mov ax,cs
	mov ds,ax

	mov ah,06H
	mov dl,service_code
	int 21H
	mov dl,sub_code
	mov ah,06H
	int 21H
	mov dx, offset debug_msg
	mov ah, 09H          ; Display greeting
	int 21H

	pop dx
	pop ax
	pop bx
	pop ds
	ret
endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;; Auxiliary Procedures ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; return 0 in ax if string pointed to by es:bx is the same as
; the one pointed to by ds:dx. String pointed to by es:bx 
; has length in cx

strncmp proc
	push di
	push si
	mov di,bx
	mov si,dx
	cld
repe	cmpsb 
	mov ax,cx
	pop si
	pop di
	ret
endp

; print number 0-65k in ax as decimal to screen
; followed by carriage return newline
; Assumes ds equals cs

convertee db 0,0,0,0,0,'$' ; buffer to hold digits
base dw 10             ; base for conversion

print_num proc

	push ax
	push bx
	push dx
	mov bx,4       ; index of digit in convertee buffer
	mov base,10
convert_cycle:
	xor dx,dx
	div base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[bx],dl
	dec bx         ; point to next digit position
	cmp ax,0
	jne convert_cycle ; keep dividing until zero
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H
	pop dx
	pop bx
	pop ax
	ret
endp

; Print version/specification for XMS
; The version is in ax,  specification in bx
; Both in BCD

XMSVERSION1 db 'XMS Version ','$'
XMSVERSION2 db ' (Specification ','$'
XMSVERSION3 db ')',0AH,0DH,'$'

print_version proc
	push ax
	push bx
	push dx

	; print start of line
	mov dx,offset XMSVERSION1
	mov ah,09H
	int 21H

; convert version code in ax to x.xx 
	mov base,16    ; repeated division by 16
	xor dx,dx      ; prepare to divide base into dx:ax
	div base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[4],dl
	xor dx,dx      ; repeat for other digits, right to left
	div base
	add dx,48
	mov convertee[3],dl
	mov convertee[2],46   ; ascii code for dot	
	xor dx,dx
	div base
	add dx,48
	mov convertee[1],dl
	mov convertee[0],32   ; ascii code for space
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H

	; set up to print specifiaction
	mov dx,offset XMSVERSION2
	mov ah,09H
	int 21H

; Now repeat conversion with bx in place of ax
	mov ax,bx
	mov base,16    ; repeated division by 16
	xor dx,dx      ; prepare to divide base into dx:ax
	div base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[4],dl
	xor dx,dx      ; repeat for other digits, right to left
	div base
	add dx,48
	mov convertee[3],dl
	mov convertee[2],46   ; ascii code for dot	
	xor dx,dx
	div base
	add dx,48
	mov convertee[1],dl
	mov convertee[0],32   ; ascii code for space
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H

	; and print end of line
	mov dx,offset XMSVERSION3
	mov ah,09H
	int 21H
	
	pop dx
	pop bx
	pop ax
	ret
endp
ram:  ; below is frame buffer
_TEXT ends
end
